<?php
  
function dnd_character_sheet_display($node) { 
  global $p;
  global $xoffset;
  global $yoffset;
  global $row;

  $p = new PDFlib();
  
  // US Letter page size in points
  $page_width     = 612.0; //540 -> 612
  $page_height    = 792.0; //720 -> 792
  
  $margin_height  = 36.0;
  $margin_width   = 36.0;
  
  $xoffset        = $margin_width;
  $yoffset        = $page_height - $margin_height - DND_CHARACTER_SHEET_LINE_HEIGHT;
  
  
  $row = array(0.0);
  
  //  open new PDF file; insert a file name to create the PDF on disk 
  if ($p->begin_document("", "destination={page=1 type=fixed zoom=1} viewerpreferences={printscaling=none centerwindow=true}") == 0) {
    drupal_set_message("Error creating PDF: " . $p->get_errmsg(), 'error');
    return FALSE;
  }
  
  // Set main PDF information
  $p->set_info("Creator", "http://dnd.tobbyhagler.com");
  $p->set_info("Author", "Tobby Hagler");
  $p->set_info("Title", "D&D Character Record Sheet");
  $p->set_info("Subject", "D&D Character Record Sheet");
  
  // Page 1
  $p->begin_page_ext($page_width, $page_height, "");
    $row = define_rows(1);

    // Display the basic information (name, race, class, etc.)
    _dnd_character_sheet_character_basic_info($node, $p, $row, $xoffset, $yoffset);

    // Display the primary stats (Ability Scores, Saves, HP, AC, Initiative, etc.)
    _dnd_character_sheet_character_main_stats($node, $row, $p);
    
    // Display the AC and other armor and protection stats
    _dnd_character_sheet_character_armor($node, $row, $p);
    
    // Display the combat stats
    _dnd_character_sheet_character_combat_stats($node, $row, $p);
    
    // Display the weapon info
    _dnd_character_sheet_character_weapons($node, $row, $p);

  $p->end_page_ext("");


  // Page 2
  $p->begin_page_ext($page_width, $page_height, "");
    $row = define_rows(2);

  // Display skill points and skill assignments
  _dnd_character_sheet_character_skills($node, $row, $p);
  
  $p->end_page_ext("");
  
  // Page 3
  $p->begin_page_ext($page_width, $page_height, "");
    $row = define_rows(3);
    // Display feats 
    _dnd_character_sheet_character_feats($node, $row, $p);
  
  $p->end_page_ext("");
  
  // Page 4
  $p->begin_page_ext($page_width, $page_height, "");
    $row = define_rows(4);

  // Display Equipment
  equipment($node);

  // Display Wealth
  wealth($node);
  
  // TODO: This PDF sheet has little error checking, so it throws lots of PHP strict warnings
  // about fields with illegal offsets for empty fields. This *CANNOT* leave alpha until this
  // is fixed.
  drupal_get_messages('error', TRUE);
  
  $p->end_page_ext("");

  $p->end_document("");
  
  // Output the PDF to the browser
  $buf = $p->get_buffer();
  $len = strlen($buf);
  
  header("Content-type: application/pdf");
  header("Content-Length: $len");
  header("Content-Disposition: inline; filename=sheet" . time() . ".pdf");
  print $buf;
}



function _dnd_character_sheet_character_basic_info($node, &$p, &$row, &$xoffset, &$yoffset) {
  // Character Name
  underline($p, 0, $row[0], 310, $row[0]);
  label_under($p, "Character Name", 0, $row[0]);
  underlined_text($p, $node->title, 0, $row[0]);

  // Race
  underline($p, 315, $row[0], 420, $row[0]);
  label_under($p, "Race", 315, $row[0]);
  underlined_text($p, $node->field_dnd_character_race[$node->language][0]['value'], 315, $row[0]);

  // Class
  underline($p, 425, $row[0], 540, $row[0]);
  label_under($p, "Class", 425, $row[0]);
  underlined_text($p, $node->field_dnd_character_class_level[$node->language][0]['class'], 425, $row[0]);

  // Alignment
  underline($p, 0, $row[1], 165, $row[1]);
  label_under($p, "Alignment", 0, $row[1]);
  underlined_text($p, $node->field_dnd_character_alignment[$node->language][0]['value'], 0, $row[1]);

  // Deity
  underline($p, 170, $row[1], 310, $row[1]);
  label_under($p, "Deity", 170, $row[1]);
  underlined_text($p, $node->field_dnd_character_deity[$node->language][0]['value'], 170, $row[1]);

  // Level
  underline($p, 315, $row[1], 385, $row[1]);
  label_under($p, "Level", 315, $row[1]);
  underlined_text($p, $node->field_dnd_character_class_level[$node->language][0]['level'], 315, $row[1]);

  // Experience
  underline($p, 390, $row[1], 460, $row[1]);
  label_under($p, "Experience", 390, $row[1]);
  underlined_text($p,$node->field_dnd_character_xp[$node->language][0]['value'], 390, $row[1]);

  // Penalty
  underline($p, 465, $row[1], 540, $row[1]);
  label_under($p, "Penalty", 465, $row[1]);
  underlined_text($p, $node->field_dnd_character_xp_penalty[$node->language][0]['value'], 465, $row[1]);

  // Gender
  underline($p, 0, $row[2], 105, $row[2]);
  label_under($p, "Gender", 0, $row[2]);
  underlined_text($p, $node->field_dnd_character_gender[$node->language][0]['value'], 0, $row[2]);
  
  // Age
  underline($p, 110, $row[2], 150, $row[2]);
  label_under($p, "Age", 110, $row[2]);
  underlined_text($p, $node->field_dnd_character_age[$node->language][0]['value'], 110, $row[2]);
  
  // Size
  underline($p, 155, $row[2], 200, $row[2]);
  label_under($p, "Size", 155, $row[2]);
  underlined_text($p, $node->field_dnd_character_size[$node->language][0]['value'], 155, $row[2]);
  
  // Height
  underline($p, 205, $row[2], 265, $row[2]);
  label_under($p, "Height", 205, $row[2]);
  underlined_text($p, $node->field_dnd_character_height[$node->language][0]['value'], 205, $row[2]);
  
  // Weight
  underline($p, 270, $row[2], 340, $row[2]);
  label_under($p, "Weight", 270, $row[2]);
  underlined_text($p, $node->field_dnd_character_weight[$node->language][0]['value'], 270, $row[2]);
  
  // Skin
  underline($p, 345, $row[2], 390, $row[2]);
  label_under($p, "Skin", 345, $row[2]);
  underlined_text($p, $node->field_dnd_character_skin[$node->language][0]['value'], 345, $row[2]);
  
  // Hair
  underline($p, 395, $row[2], 430, $row[2]);
  label_under($p, "Hair", 395, $row[2]);
  underlined_text($p, $node->field_dnd_character_hair[$node->language][0]['value'], 395, $row[2]);
  
  // Eyes
  underline($p, 435, $row[2], 490, $row[2]);
  label_under($p, "Eyes", 435, $row[2]);
  underlined_text($p, $node->field_dnd_character_eyes[$node->language][0]['value'], 435, $row[2]);
  
  // Laterality
  underline($p, 495, $row[2], 540, $row[2]);
  label_under($p, "Laterality", 495, $row[2]);
  underlined_text($p, $node->field_dnd_character_laterality[$node->language][0]['value'], 495, $row[2]);
  
  // Biography
  $text = explode("|||", wordwrap(strip_tags($node->body[$node->language][0]['value']), 80, '|||'));
  underline($p, 0, $row[3], 320, $row[3]);
  underline($p, 0, $row[4], 320, $row[4]);
  underline($p, 0, $row[5], 320, $row[5]);
  underlined_text($p, $text[0], 0, $row[3], '', 'small');
  underlined_text($p, $text[1], 0, $row[4], '', 'small');
  underlined_text($p, $text[2], 0, $row[5], '', 'small');
  label_under($p, "Biography", 0, $row[5]);
  
  // History
  $text = explode("|||", wordwrap($node->field_dnd_character_history[$node->language][0]['value'], 54, '|||'));
  underline($p, 325, $row[3], 540, $row[3]);
  underline($p, 325, $row[4], 540, $row[4]);
  underline($p, 325, $row[5], 541, $row[5]);
  underlined_text($p, $text[0], 325, $row[3], '', 'small');
  underlined_text($p, $text[1], 325, $row[4], '', 'small');
  underlined_text($p, $text[2], 325, $row[5], '', 'small');
  label_under($p, "History", 325, $row[5]);
  
  // Goals
  $text = explode("|||", wordwrap($node->field_dnd_character_goals[$node->language][0]['value'], 80, '|||'));
  underline($p, 0, $row[6], 320, $row[6]);
  underline($p, 0, $row[7], 320, $row[7]);
  underline($p, 0, $row[8], 320, $row[8]);
  underlined_text($p, $text[0], 0, $row[6], '', 'small');
  underlined_text($p, $text[1], 0, $row[7], '', 'small');
  underlined_text($p, $text[2], 0, $row[8], '', 'small');
  label_under($p, "Goals", 0, $row[8]);
  
  // Appearance
  $text = explode("|||", wordwrap($node->field_dnd_character_appearance[$node->language][0]['value'], 54, '|||'));
  underline($p, 325, $row[6], 540, $row[6]);
  underline($p, 325, $row[7], 540, $row[7]);
  underline($p, 325, $row[8], 541, $row[8]);
  underlined_text($p, $text[0], 325, $row[6], '', 'small');
  underlined_text($p, $text[1], 325, $row[7], '', 'small');
  underlined_text($p, $text[2], 325, $row[8], '', 'small');
  label_under($p, "Appearance", 325, $row[8]);
  
  // Personality/Behavior
  $text = explode("|||", wordwrap($node->field_dnd_character_personality[$node->language][0]['value'], 80, '|||'));
  underline($p, 0, $row[9], 320, $row[9]);
  underline($p, 0, $row[10], 320, $row[10]);
  underline($p, 0, $row[11], 320, $row[11]);
  underlined_text($p, $text[0], 0, $row[9], '', 'small');
  underlined_text($p, $text[1], 0, $row[10], '', 'small');
  underlined_text($p, $text[2], 0, $row[11], '', 'small');
  label_under($p, "Personality/Behavior", 0, $row[11]);
  
  // Notes
  $text = explode("|||", wordwrap($node->field_dnd_character_notes[$node->language][0]['value'], 54, '|||'));
  underline($p, 325, $row[9], 540, $row[9]);
  underline($p, 325, $row[10], 540, $row[10]);
  underline($p, 325, $row[11], 541, $row[11]);
  for ($i = 0; $i < 3; $i ++) {
    if (!empty($text[$i])) underlined_text($p, $text[$i], 325, $row[9 + $i], '', 'small');
  }
  label_under($p, "Notes", 325, $row[11], "");
}

function _dnd_character_sheet_character_main_stats($node, &$row, &$p) {
  // Ability Scores
  label_filled($p, "Ability Scores", 0, $row[12], 125);
  label_over($p, "Score", 30, $row[13] + 9, 23);
  label_over($p, "Mod", 53, $row[13] + 9, 22);
  label_over($p, "Temp", 79, $row[13] + 2, 22);
  label_over($p, "Mod", 79, $row[13] + 9, 22);
  label_over($p, "Temp", 102, $row[13] + 2, 23);
  label_over($p, "Mod", 102, $row[13] + 9, 23);
  $abilities = array("str", "dex", "con", "int", "wis", "chr");
  for ($i = 0; $i < 6; $i ++) {
    label_filled($p, $abilities[$i], 0, $row[$i + 14], 25);
      cell($p, $node->field_dnd_character_abilities[$node->language][$i]['score'], 30, $row[$i + 14], 23);
      $value = ($node->field_dnd_character_abilities[$node->language][$i]['mod'] > 0) 
        ? '+' . $node->field_dnd_character_abilities[$node->language][$i]['mod'] 
        : $node->field_dnd_character_abilities[$node->language][$i]['mod'];
      cell($p, $value, 53, $row[$i + 14], 22);
      
      $value = ($node->field_dnd_character_abilities[$node->language][$i]['tempscore'] > 0) 
        ? '+' . $node->field_dnd_character_abilities[$node->language][$i]['tempscore'] : '';
      cell($p, $value, 79, $row[$i + 14], 23, '', 'dashed');
      
      $value = ($node->field_dnd_character_abilities[$node->language][$i]['tempmod'] > 0) 
        ? '+' . $node->field_dnd_character_abilities[$node->language][$i]['tempmod'] : '';
      cell($p, $value, 102, $row[$i + 14], 22, '', 'dashed');
  }
  
  // Saving Throws
  label_filled($p, "Saving Throws", 130, $row[12], 220);
  label_over($p, "Total", 179, $row[13] + 9, 30);
  label_over($p, "Base", 208, $row[13] + 2, 30);
  label_over($p, "Save", 208, $row[13] + 9, 30);
  label_over($p, "Abilty", 236, $row[13] + 2, 30);
  label_over($p, "Mod", 236, $row[13] + 9, 30);
  label_over($p, "Magic", 264, $row[13] + 2, 30);
  label_over($p, "Mod", 264, $row[13] + 9, 30);
  label_over($p, "Misc", 292, $row[13] + 2, 30);
  label_over($p, "Mod", 292, $row[13] + 9, 30);
  label_over($p, "Temp", 321, $row[13] + 2, 30);
  label_over($p, "Score", 321, $row[13] + 9, 30);
  
  $labels = array('fort', 'ref', 'will');
  $mods   = array('(Con)', '(Dex)', '(Wis)');

  for ($i = 0; $i < 3; $i ++) {
    label_filled($p, $labels[$i], 130, $row[$i + 14], 30);
    label_middle($p, $mods[$i], 160, $row[$i + 14], 20);

    // total save
    cell($p, $node->field_dnd_character_saves[$node->language][$i]['total'], 180, $row[$i + 14], 27);
    label_middle($p, "+", 207, $row[$i + 14], 5);

    // base save
    cell($p, $node->field_dnd_character_saves[$node->language][$i]['base'], (212 + (0 * 28)), $row[$i + 14], 22);
    label_middle($p, "+", (234 + (0 * 28)), $row[$i + 14], 6);
    
    // ability mod
    cell($p, $node->field_dnd_character_saves[$node->language][$i]['abilitymod'], (212 + (1 * 28)), $row[$i + 14], 22);
    label_middle($p, "+", (234 + (1 * 28)), $row[$i + 14], 6);
    
    // magic mod
    cell($p, $node->field_dnd_character_saves[$node->language][$i]['magicmod'], (212 + (2 * 28)), $row[$i + 14], 22);
    label_middle($p, "+", (234 + (2 * 28)), $row[$i + 14], 6);
    
    // misc mod
    cell($p, $node->field_dnd_character_saves[$node->language][$i]['miscmod'], (212 + (3 * 28)), $row[$i + 14], 22);
    label_middle($p, "+", (234 + (3 * 28)), $row[$i + 14], 6);
    
    label_middle($p, "+", 318, $row[$i + 14], 6);
    cell($p, $node->field_dnd_character_saves[$node->language][$i]['temp'], 324, $row[$i + 14], 25, "", "dashed");
  }
  label_over($p, "Saving Throw Modifiers", 130, $row[17] + 4, 220);
  cell($p, $node->field_dnd_character_save_mods[$node->language][0]['value'], 130, $row[18], 220);
  cell($p, $node->field_dnd_character_save_mods[$node->language][1]['value'], 130, $row[19], 220);
          
  // Hit Points
  label_filled($p, "Hit Points", 355, $row[12], 125);
  cell($p, str_pad($node->field_dnd_character_hp[$node->language][0]['value'], 14), 480, $row[12], 60);
  divider($p, 512, $row[12] + 2, 508, $row[12] + 13);

  // Encumberance
  $capacity = _dnd_character_sheet_get_carry_load($node);
  label_over($p, "Lift Overhead", 385, $row[14] - 3, 30);
  label_over($p, "(1xMax Load)", 385, $row[14] + 4, 30);
  cell($p, $capacity['heavy']['weight'], 385, $row[15], 30);
  
  label_over($p, "Lift Off Ground", 435, $row[14] - 3, 30);
  label_over($p, "(2xMax Load)", 435, $row[14] + 4, 30);
  cell($p, intval($capacity['heavy']['weight']) * 2, 435, $row[15], 30);
  
  label_over($p, "Push/Drag", 485, $row[14] - 3, 30);
  label_over($p, "(5xMax Load)", 485, $row[14] + 4, 30);
  cell($p, intval($capacity['heavy']['weight']) * 5, 485, $row[15], 30);

  label_over($p, "Load", 355, $row[16] + 4, 30);
  cell($p, "", 355, $row[17], 30);
  
  label_over($p, "Speed", 355, $row[18] + 4, 30);
  cell($p, $capacity['light']['speed'], 355, $row[19], 30);
  
  label_over($p, "Speed", 410, $row[16] + 4, 26);
  label_over($p, "Max", 436, $row[16] -3, 26);
  label_over($p, "Load", 436, $row[16] + 4, 26);
  label_over($p, "Max", 462, $row[16] -3, 26);
  label_over($p, "Dex", 462, $row[16] + 4, 26);
  label_over($p, "Check", 488, $row[16] - 3, 26);
  label_over($p, "Pen", 488, $row[16] + 4, 26);
  label_over($p, "Run", 514, $row[16] + 4, 26);
  $encumberances = array('L', 'M', 'H');
  
  $i = 0;
  label_middle($p, $encumberances[$i], 402, $row[$i + 17], 8);
  cell($p, $capacity['light']['speed'], 26 * 0 + 410, $row[$i + 17], 26);
  cell($p, $capacity['light']['weight'], 26 * 1 + 410, $row[$i + 17], 26);
  cell($p, '-', 26 * 0 + 462, $row[$i + 17], 26, '', 'dashed');
  cell($p, '-', 26 * 1 + 462, $row[$i + 17], 26, '', 'dashed');
  cell($p, intval($capacity['light']['speed']) * 4, 26 * 2 + 462, $row[$i + 17], 26, '', 'dashed');
  
  $i = 1;
  label_middle($p, $encumberances[$i], 402, $row[$i + 17], 8);
  cell($p, $capacity['medium']['speed'], 26 * 0 + 410, $row[$i + 17], 26);
  cell($p, $capacity['medium']['weight'], 26 * 1 + 410, $row[$i + 17], 26);
  cell($p, $capacity['medium']['max_dex'], 26 * 0 + 462, $row[$i + 17], 26, '', 'dashed');
  cell($p, $capacity['medium']['check_penalty'], 26 * 1 + 462, $row[$i + 17], 26, '', 'dashed');
  cell($p, intval($capacity['medium']['speed']) * 4, 26 * 2 + 462, $row[$i + 17], 26, '', 'dashed');
  
  $i = 2;
  label_middle($p, $encumberances[$i], 402, $row[$i + 17], 8);
  cell($p, $capacity['heavy']['speed'], 26 * 0 + 410, $row[$i + 17], 26);
  cell($p, $capacity['heavy']['weight'], 26 * 1 + 410, $row[$i + 17], 26);
  cell($p, $capacity['heavy']['max_dex'], 26 * 0 + 462, $row[$i + 17], 26, '', 'dashed');
  cell($p, $capacity['heavy']['check_penalty'], 26 * 1 + 462, $row[$i + 17], 26, '', 'dashed');
  cell($p, intval($capacity['heavy']['speed']) * 3, 26 * 2 + 462, $row[$i + 17], 26, '', 'dashed');
}

function _dnd_character_sheet_character_armor($node, &$row, &$p) {
  // Armor Class
  label_filled($p, "Armor Class", 0, $row[20], 85);
  cell($p, $node->field_dnd_character_armor_bonus[$node->language][0]['value'], 85, $row[20], 22);
  label_over($p, "Normal", 85, $row[21] - 5, 22);
  label_middle($p, '=', 107, $row[20], 6);
  cell($p, '10', 113, $row[20], 22);
  $ac_modifiers1 = array("Armor", "Shield", "Dex Mod", "Size Mod", "Misc Mod");
  $ac_modifiers2 = array("Bonus", "Bonus", "", "", "");
  
  $i = 0;
  label_middle($p, '+', $i * 28 + 135, $row[20], 6);
  cell($p, $node->field_dnd_character_armor_bonus[$node->language][0]['value'], $i * 28 + 135 + 6, $row[20], 22);
  label_over($p, $ac_modifiers1[$i], $i * 28 + 135 + 6, $row[21] - 5, 22);
  label_over($p, $ac_modifiers2[$i], $i * 28 + 135 + 6, $row[21] + 1, 22);
  
  $i = 1;
  label_middle($p, '+', $i * 28 + 135, $row[20], 6);
  cell($p, $node->field_dnd_character_shield_bon[$node->language][0]['value'], $i * 28 + 135 + 6, $row[20], 22);
  label_over($p, $ac_modifiers1[$i], $i * 28 + 135 + 6, $row[21] - 5, 22);
  label_over($p, $ac_modifiers2[$i], $i * 28 + 135 + 6, $row[21] + 1, 22);
  
  $i = 2;
  label_middle($p, '+', $i * 28 + 135, $row[20], 6);
  cell($p, $node->field_dnd_character_abilities[$node->language][1]['mod'], $i * 28 + 135 + 6, $row[20], 22);
  label_over($p, $ac_modifiers1[$i], $i * 28 + 135 + 6, $row[21] - 5, 22);
  label_over($p, $ac_modifiers2[$i], $i * 28 + 135 + 6, $row[21] + 1, 22);
  
  $i = 3;
  label_middle($p, '+', $i * 28 + 135, $row[20], 6);
  cell($p, $node->field_dnd_character_ac_size_mod[$node->language][0]['value'], $i * 28 + 135 + 6, $row[20], 22);
  label_over($p, $ac_modifiers1[$i], $i * 28 + 135 + 6, $row[21] - 5, 22);
  label_over($p, $ac_modifiers2[$i], $i * 28 + 135 + 6, $row[21] + 1, 22);
  
  $i = 4;
  label_middle($p, '+', $i * 28 + 135, $row[20], 6);
  cell($p, $node->field_dnd_character_ac_misc_modifier[$node->language][0]['value'], $i * 28 + 135 + 6, $row[20], 22);
  label_over($p, $ac_modifiers1[$i], $i * 28 + 135 + 6, $row[21] - 5, 22);
  label_over($p, $ac_modifiers2[$i], $i * 28 + 135 + 6, $row[21] + 1, 22);

  cell($p, $node->field_dnd_character_arcane_fail[$node->language][0]['value'], 20, $row[22], 25);
  label_over($p, 'Arcane', 20, $row[23] - 5, 25);
  label_over($p, 'Spell Fail', 20, $row[23] + 1, 25);
  
  cell($p, $node->field_dnd_character_ac_flat_foot[$node->language][0]['value'], 65, $row[22], 25);
  label_over($p, 'Flat Foot', 65, $row[23] - 5, 25);
  
  cell($p, $node->field_dnd_character_ac_check_pen[$node->language][0]['value'], 20, $row[24], 25);
  label_over($p, 'Armor', 20, $row[25] - 5, 25);
  label_over($p, 'Check Pen', 20, $row[25] + 1, 25);
  
  cell($p, $node->field_dnd_character_ac_vs_touch[$node->language][0]['value'], 65, $row[24], 25);
  label_over($p, 'Vs Touch', 65, $row[25] - 5, 25);
  
  label_over($p, 'AC Modifiers', 107, $row[22] + 5, 168);
  cell($p, $node->field_dnd_character_ac_modifiers[$node->language][0]['value'], 107, $row[23], 168);
  cell($p, $node->field_dnd_character_ac_modifiers[$node->language][1]['value'], 107, $row[24], 168);
  cell($p, $node->field_dnd_character_ac_modifiers[$node->language][2]['value'], 107, $row[25], 168);
  
  // Armor and Protection
  label_filled($p, "Armor & Protection", 280, $row[20], 120);
  /*Type  	Armor Bonus  	Max Dex  	Chk Pen  	Spell Fail  	Speed  	Wgt Lbs*/
  label_over($p, 'Type', 400, $row[20] + 5, 20);
  label_over($p, 'Armor', 420, $row[20] - 1, 20);
  label_over($p, 'Bonus', 420, $row[20] + 5, 20);
  label_over($p, 'Max', 440, $row[20] - 1, 20);
  label_over($p, 'Dex', 440, $row[20] + 5, 20);
  label_over($p, 'Check', 460, $row[20] -1, 20);
  label_over($p, 'Pen', 460, $row[20] + 5, 20);
  label_over($p, 'Spell', 480, $row[20] - 1, 20);
  label_over($p, 'Fail', 480, $row[20] + 5, 20);
  label_over($p, 'Speed', 500, $row[20] + 5, 20);
  label_over($p, 'Wgt', 520, $row[20] - 1, 20);
  label_over($p, 'Lbs', 520, $row[20] + 5, 20);
  
  for($i = 0; $i < 5; $i ++) {
    $node->field_dnd_character_armor[$node->language][$i]['type'] = ''; // Not used; set here to keep from throwing strict warnings
    cell($p, $node->field_dnd_character_armor[$node->language][$i]['armor'], 280, $row[21 + $i], 120);
    cell($p, $node->field_dnd_character_armor[$node->language][$i]['type'], 0 * 20 + 400, $row[21 + $i], 20);
    cell($p, $node->field_dnd_character_armor[$node->language][$i]['bonus'], 1 * 20 + 400, $row[21 + $i], 20);
    cell($p, $node->field_dnd_character_armor[$node->language][$i]['max_dex'], 2 * 20 + 400, $row[21 + $i], 20);
    cell($p, $node->field_dnd_character_armor[$node->language][$i]['check_pen'], 3 * 20 + 400, $row[21 + $i], 20);
    cell($p, $node->field_dnd_character_armor[$node->language][$i]['spell_fail'], 4 * 20 + 400, $row[21 + $i], 20);
    cell($p, $node->field_dnd_character_armor[$node->language][$i]['speed'], 5 * 20 + 400, $row[21 + $i], 20);
    cell($p, $node->field_dnd_character_armor[$node->language][$i]['weight'], 6 * 20 + 400, $row[21 + $i], 20);
  }
}

function _dnd_character_sheet_character_combat_stats($node, &$row, &$p) {
  // Initiative
  label_filled($p, "Initiative", 0, $row[26], 80);
  cell($p, $node->field_dnd_character_init[$node->language][0]['value'], 80, $row[26], 20);
  label_middle($p, "=", 100, $row[26], 6);
  cell($p, $node->field_dnd_character_abilities[$node->language][1]['mod'], 106, $row[26], 20);
  label_middle($p, "+", 126, $row[26], 6);
  cell($p, $node->field_dnd_character_init_mod[$node->language][0]['value'], 132, $row[26], 20);
  label_over($p, "Dex", 106, $row[27] - 5, 20);
  label_over($p, "Mod", 106, $row[27] + 1, 20);
  label_over($p, "Misc", 132, $row[27] - 5, 20);
  label_over($p, "Mod", 132, $row[27] + 1, 20);
  label_over($p, 'Initiative Modifiers', 0, $row[27] + 5, 152);
  for ($i = 0; $i < 3; $i ++) {
      cell($p, $node->field_dnd_character_init_mods[$node->language][$i]['value'], 0, $row[28 + $i], 152);
  }

  // Attack Bonuses
  label_filled($p, "Attack Bonuses", 157, $row[26], 105);

  // Base Attack Bonus
  label_filled($p, "Base Attack Bonus", 267, $row[26], 130);
  cell($p, $node->field_dnd_character_base_attack[$node->language][0]['value'], 397, $row[26], 23);
  
  // Attacks
  label_over($p, 'Total', 191, $row[27] + 5, 152);
  label_over($p, 'Base', 222, $row[27] - 1, 152);
  label_over($p, 'Attack', 222, $row[27] + 5, 152);
  label_over($p, 'Ability', 249, $row[27] - 1, 152);
  label_over($p, 'Mod',     249, $row[27] + 5, 152);
  label_over($p, 'Size',  276, $row[27] - 1, 152);
  label_over($p, 'Mod',   276, $row[27] + 5, 152);
  label_over($p, 'Misc', 305, $row[27] -1, 152);
  label_over($p, 'Mod',  305, $row[27] + 5, 152);
  
  $size_mod = ($node->field_dnd_character_size[$node->language][0]['value'] == 'small') ? 2 : 0;
  
  label_filled($p, "Melee", 157, $row[28], 74);
  label_middle($p, "(Str)", 231, $row[28], 25);

  $i = 0;
  cell($p, $node->field_dnd_character_melee[$node->language][0]['value'], 256, $row[28 + $i], 22);
  label_middle($p, '=', 278, $row[28 + $i], 8);
    $j = 0;
    cell($p, $node->field_dnd_character_base_attack[$node->language][0]['value'], $j * 28 + 286, $row[28 + $i], 22);
    label_middle($p, '+', $j * 28 + 308, $row[28 + $i], 6);

    $j = 1;
    cell($p, $node->field_dnd_character_abilities[$node->language][0]['mod'], $j * 28 + 286, $row[28 + $i], 22);
    label_middle($p, '+', $j * 28 + 308, $row[28 + $i], 6);

    $j = 2;
    cell($p, $size_mod, $j * 28 + 286, $row[28 + $i], 22);
    label_middle($p, '+', $j * 28 + 308, $row[28 + $i], 6);

    $j = 3;
    cell($p, '', $j * 28 + 286, $row[28 + $i], 22);
    label_middle($p, '+', $j * 28 + 308, $row[28 + $i], 6);
  cell($p, '', 4 * 28 + 286, $row[28 + $i], 22, '', 'dashed');
  

  label_filled($p, "Ranged", 157, $row[29], 74);
  label_middle($p, "(Dex)", 231, $row[29], 25);
  $i = 1;
  cell($p, $node->field_dnd_character_ranged[$node->language][0]['value'], 256, $row[28 + $i], 22);
  label_middle($p, '=', 278, $row[28 + $i], 8);
    $j = 0;
    cell($p, $node->field_dnd_character_base_attack[$node->language][0]['value'], $j * 28 + 286, $row[28 + $i], 22);
    label_middle($p, '+', $j * 28 + 308, $row[28 + $i], 6);

    $j = 1;
    cell($p, $node->field_dnd_character_abilities[$node->language][1]['mod'], $j * 28 + 286, $row[28 + $i], 22);
    label_middle($p, '+', $j * 28 + 308, $row[28 + $i], 6);

    $j = 2;
    cell($p, $size_mod, $j * 28 + 286, $row[28 + $i], 22);
    label_middle($p, '+', $j * 28 + 308, $row[28 + $i], 6);

    $j = 3;
    cell($p, '', $j * 28 + 286, $row[28 + $i], 22);
    label_middle($p, '+', $j * 28 + 308, $row[28 + $i], 6);
  cell($p, '', 4 * 28 + 286, $row[28 + $i], 22, '', 'dashed');

  label_filled($p, "", 157, $row[30], 74);
  label_middle($p, "", 231, $row[30], 25);
    $i = 2;
    cell($p, '', 256, $row[28 + $i], 22);
    label_middle($p, '=', 278, $row[28 + $i], 8);
    for ($j = 0; $j < 4; $j ++) {
      cell($p, '', $j * 28 + 286, $row[28 + $i], 22);
      label_middle($p, '+', $j * 28 + 308, $row[28 + $i], 6);
    }
    cell($p, '', 4 * 28 + 286, $row[28 + $i], 22, '', 'dashed');

  // Attack Bonus Modifiers
  label_over($p, "Attack Bonus Modifiers", 425, $row[26] + 5, 115);
  for ($i = 0; $i < 4; $i ++) {
    cell($p, $node->field_dnd_character_attack_mods[$node->language][$i]['value'], 425, $row[27 + $i], 115);
  }
}

function _dnd_character_sheet_character_weapons($node, &$row, &$p) {
  label_filled($p, "Weapons", 0, $row[31], 145);
  $weapon_widths = array(145, 60, 45, 40, 40, 25, 25, 25, 135);
  $weapon_labels = array("", "Attack Bonus", "Damage", "Critical", "Range", "Size", "Type", "Weight", "Notes & Related Feats");
  $current_width = 145;
  for ($i = 1; $i < 9; $i ++) {
    label_over($p, $weapon_labels[$i], $current_width, $row[31] + 4, $weapon_widths[$i]);
    $current_width += $weapon_widths[$i];
  }
  for ($i = 0; $i < 7; $i ++) {
    $current_width = 0;
    
    $j =0;
    cell($p, $node->field_dnd_character_weapons[$node->language][$i]['weapon'], $current_width, $row[32 + $i], $weapon_widths[$j], '', '', '', 'left');
    $current_width += $weapon_widths[$j];
    
    $j = 1;
    /*
    $attack_bonus = ($node->character['weapons'][$i])
      ? ($node->character['weapons'][$i]['subcategory'] == 'Ranged Weapons') 
        ? $node->character['base_attack_bonuses']['ranged'] 
        : $node->character['base_attack_bonuses']['melee'] 
      : '';
    */
    cell($p, $node->field_dnd_character_weapons[$node->language][$i]['bonus'], $current_width, $row[32 + $i], $weapon_widths[$j]);
    $current_width += $weapon_widths[$j];
    
    $j = 2;
    cell($p, $node->field_dnd_character_weapons[$node->language][$i]['damage'], $current_width, $row[32 + $i], $weapon_widths[$j]);
    $current_width += $weapon_widths[$j];
    
    $j = 3;
    cell($p, $node->field_dnd_character_weapons[$node->language][$i]['critical'], $current_width, $row[32 + $i], $weapon_widths[$j]);
    $current_width += $weapon_widths[$j];
    
    $j = 4;
    cell($p, $node->field_dnd_character_weapons[$node->language][$i]['range'], $current_width, $row[32 + $i], $weapon_widths[$j]);
    $current_width += $weapon_widths[$j];
    
    $j = 5;
    cell($p, $node->field_dnd_character_weapons[$node->language][$i]['size'], $current_width, $row[32 + $i], $weapon_widths[$j]);
    $current_width += $weapon_widths[$j];
    
    $j = 6;
    cell($p, $node->field_dnd_character_weapons[$node->language][$i]['type'], $current_width, $row[32 + $i], $weapon_widths[$j]);
    $current_width += $weapon_widths[$j];
    
    $j = 7;
    cell($p, $node->field_dnd_character_weapons[$node->language][$i]['weight'], $current_width, $row[32 + $i], $weapon_widths[$j]);
    $current_width += $weapon_widths[$j];
    
    $j = 8;
    cell($p, $node->field_dnd_character_weapons[$node->language][$i]['notes'], $current_width, $row[32 + $i], $weapon_widths[$j], '', '', '', 'left');
    $current_width += $weapon_widths[$j];
  }
  
  // Weapon Proficiencies
  label_filled($p, "Weapon Proficiencies", 0, $row[39], 155);
  for ($i = 0; $i < 3; $i ++) {
    cell($p, $node->character_weapon_prof[$node->language][$i]['value'], 0, $row[40 + $i], 240);
  }
  
  // Combat Notes
  label_filled($p, "Combat Notes", 245, $row[39], 295);
  for ($i = 0; $i < 3; $i ++) {
    cell($p, $node->character_combat_notes[$node->language][$i]['value'], 245, $row[40 + $i], 295);
  }
}

function _dnd_character_sheet_character_skills($node, &$row, &$p) {
  $skill_data = $node->character['skills'];

  $top_class = array_shift($node->character['class']['full']);
  //print '<pre>' . print_r($top_class['skill_points'], true);
  //print '<pre>' . print_r($skill_data, true);
  //die;

  label_filled($p, "Skills", 0, $row[0], 50);
  label_over($p, "Skill Points Per Level", 60, $row[0], 60);
  cell($p, $top_class['skill_points'], 120, $row[0], 30);
  label_over($p, "+ Int Mod", 150, $row[0], 28);
  label_over($p, "Class Skill Max Ranks = Character Level + 3", 206, $row[0] - 5, 135);
  label_over($p, "Cross-Class Skill Max Ranks = (Character Level + 3) / 2", 206, $row[0] + 3, 135);
  label_over($p, chr(149) . " Can Be Used Untrained", 376, $row[0], 60);
  label_over($p, "+ Armor Check Penalty Applies", 466, $row[0], 60);

  for ($i = 0; $i < 2; $i ++) {
    $startx = ($i % 2) ? 276 : 0;
    label_over($p, "Skill", 115 + $startx, $row[1] - 1.0, 24);
    label_over($p, "Modifier", 115 + $startx, $row[1] + 5.0, 24);
    label_over($p, "Ranks", 145 + $startx, $row[1] + 5.0, 24);
    label_over($p, "Ability", 175 + $startx, $row[1] - 1.0, 24);
    label_over($p, "Modifier", 175 + $startx, $row[1] + 5.0, 24);
    label_over($p, "Misc", 205 + $startx, $row[1] - 1.0, 24);
    label_over($p, "Modifier", 205 + $startx, $row[1] + 5.0, 24);
    label_over($p, "Key", 229 + $startx, $row[1] - 1.0, 20);
    label_over($p, "Ability", 229 + $startx, $row[1] + 5.0, 20);
    label_over($p, "Cross", 253 + $startx, $row[1] - 1.0, 10);
    label_over($p, "Class", 253 + $startx, $row[1] + 5.0, 10);
  }

  $skill_count = 0;
  for ($i = 0; $i < 2; $i ++) {
  	for ($j = 0; $j < 23; $j ++) {
    	$startx = ($i) ? 276 : 0;
    	$skill_string = '';
    	$skill_ability = '';
    	if (isset($skill_data[$skill_count])) {
    		if ($skill_count < (count($skill_data) * ($i + 1) / 2) ) {
    		  $skill = array(
            'name'      => $skill_data[$skill_count]['name'],
            'ability'   => strtoupper($skill_data[$skill_count]['key_ability']),
            'ranks'     => $skill_data[$skill_count]['ranks'],
            'modifier'  => $node->character['ability_bonus'][strtolower($skill_data[$skill_count]['key_ability'])],
            'score'     => intval($node->character['ability_bonus'][strtolower($skill_data[$skill_count]['key_ability'])]),
            'usage'     => ($skill_data[$skill_count]['usage'] == 'cross-class') ? 'X' : '',
            'misc_bonus'=> $skill_data[$skill_count]['misc_bonus'],
          );
          $skill['name'] .= ($skill_data[$skill_count]['trained'] == 'Yes') ? ' ' . chr(149) : '';
          $skill['name'] .= ($skill_data[$skill_count]['armor_check'] == 'Yes') ? ' +' : '';
          $skill['score']+= ($skill['usage'] == 'X') ? intval($skill_data[$skill_count]['ranks']) / 2 : intval($skill_data[$skill_count]['ranks']);
          $skill['score'] = floor($skill['score']) + intval($skill['misc_bonus']);
  			} else {
  			  $skill = array();
  				$skill_count --;
  			}
  		} else {
  		  $skill = array();
  		}
    	
    	label_skill($p, $skill['name'], 0 + $startx, $row[2 + $j], 115);
    	cell($p, $skill['score'], 115 + $startx, $row[2 + $j], 24);
        label_middle($p, '=', 139 + $startx, $row[2 + $j], 6);

        underline($p, 0 * 30 + 145 + $startx, $row[2 + $j] + 13.0, 0 * 30 + 145 + $startx + 24, $row[2 + $j] + 13.0);
        label_middle($p, '+', 0 * 30 + 145 + $startx + 24, $row[2 + $j], 6);
        underlined_text($p, $skill['ranks'], 0 * 30 + 145 + $startx + 8, $row[3 + $j] - 6);

        underline($p, 1 * 30 + 145 + $startx, $row[2 + $j] + 13.0, 1 * 30 + 145 + $startx + 24, $row[2 + $j] + 13.0);
        label_middle($p, '+', 1 * 30 + 145 + $startx + 24, $row[2 + $j], 6);
        underlined_text($p, $skill['modifier'], 1 * 30 + 145 + $startx + 8, $row[3 + $j] - 6);
        
        underline($p, 2 * 30 + 145 + $startx, $row[2 + $j] + 13.0, 2 * 30 + 145 + $startx + 24, $row[2 + $j] + 13.0);
        underlined_text($p, $skill['misc_bonus'], 2 * 30 + 145 + $startx + 8, $row[3 + $j] - 6);

        underline($p, 2 * 30 + 145 + $startx, $row[2 + $j] + 13.0, 2 * 30 + 145 + $startx + 24, $row[2 + $j] + 13.0);
        label_middle($p, $skill['ability'], 2 * 30 + 145 + $startx + 24, $row[2 + $j], 20);
        
        checkbox($p, 2 * 30 + 145 + $startx + 24 + 25, $row[2 + $j]);
        label_middle($p, $skill['usage'], 2 * 30 + 145 + $startx + 24 + 19, $row[2 + $j], 20);
    	$skill_count ++;
    }
  }

  // Modifiers and other skills
  label_filled($p, 'Skill Modifiers', 0, $row[25], 160);
  label_filled($p, 'Languages', 165, $row[25], 160);
  label_filled($p, 'Racial Traits & Class Features', 330, $row[25], 210);

  $skills_list  = $skill_data['skills_list_view'];
  $languages    = $skill_data['bonus_languages_view'];
  $traits_list  = _get_traits_list($node);
    
  for ($i = 0; $i < 16; $i ++) {
    cell($p, $skills_list[$i], 0, $row[26 + $i], 160, '', '', 'small', 'left');
    cell($p, $languages[$i], 165, $row[26 + $i], 160, '', '', 'small', 'left');
    cell($p, $traits_list[$i], 330, $row[26 + $i], 210, '', '', 'small', 'left');
  }
}


function _dnd_character_sheet_character_feats($node, &$row, &$p) {
  //$feats_list   = _get_feats_list($node);

  label_filled($p, 'Feats & Benefits', 0, $row[0], 540);
  for ($i = 0; $i < 43; $i ++) {
    cell($p, $node->field_dnd_character_feats[$node->language][$i]['value'], 0, $row[$i + 1], 540, '', '', 'small', 'left');
  }
}


function equipment($node) {
  global $row;
  global $p;

  $labels = array('Worn Equipment', 'Container', 'Container');
  for ($i = 0; $i < 17; $i ++) {
    for ($j = 0; $j < 3; $j ++) {
      $first = ($j) ? 0 : 5;
      $notfirst = ($j) ? 5 : 0;
      if ($i) {
        cell($p, '', $j * 180 + $notfirst, $row[$i], 150 + $first);
        cell($p, '', $j * 180 + 155, $row[$i], 25);
      } else {
        if ($j) {
          label_filled($p, $labels[$j], $j * 180 + $notfirst, $row[0], 60 + $first);
          cell($p, '', $j * 180 + $notfirst + 60, $row[0], 90 + $first);
          label_over($p, 'Weight', $j * 180 + 155, $row[0] + 5.0, 25);
        } else {
          label_filled($p, $labels[$j], $j * 180 + $notfirst, $row[0], 150 + $first);
          label_over($p, 'Weight', $j * 180 + 155, $row[0] + 5.0, 25); 
        }

      }
    }
  }
}

function wealth() {
  global $row;
  global $p;
  
  $labels = array('Wealth & Valuables', 'Other Valuables');
  for ($i = 0; $i < 10; $i ++) {
    for ($j = 0; $j < 2; $j ++) {
      $first = ($j) ? 0 : 5;
      $notfirst = ($j) ? 5 : 0;
      if ($i) {
        cell($p, '', $j * 270 + $notfirst, $row[17 + $i], 190 + $first);
        cell($p, '', $j * 270 + 195, $row[17 + $i], 50);
        cell($p, '', $j * 270 + 245, $row[17 + $i], 25);
      } else {
        label_filled($p, $labels[$j], $j * 270 + $notfirst, $row[17], 190 + $first);
        label_over($p, 'Location', $j * 270 + 195, $row[17] + 5.0, 50); 
        label_over($p, 'Weight', $j * 270 + 245, $row[17] + 5.0, 25); 
      }
    }
  }
}









function underline($p, $startx, $starty, $endx, $endy) {
  // Set globals
  global $xoffset;
  global $yoffset;
  
  
  // Initialize line style
  $color = rgb2cmyk(hex2rgb('#666666'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  $p->setlinewidth(DND_CHARACTER_SHEET_UNDERLINE_WIDTH);

  // And finally, draw the damn line
  $p->moveto($xoffset + $startx,  $yoffset - $starty - (DND_CHARACTER_SHEET_UNDERLINE_WIDTH / 2) - 2.0);
  $p->lineto($xoffset + $endx, $yoffset - $endy - (DND_CHARACTER_SHEET_UNDERLINE_WIDTH / 2) - 2.0);
  $p->stroke();
}

function underlined_text($p, $text, $x, $y, $options = '', $size = '') {
  // Set globals
  global $xoffset;
  global $yoffset;
  
  
  $text_size = ($size == 'small') ? DND_CHARACTER_SHEET_UNDERLINE_TEXT_SMALL : DND_CHARACTER_SHEET_UNDERLINE_TEXT;
  
  // Initialize text style
  $font = $p->load_font("Helvetica-Bold", "winansi", "");
  $p->setfont($font, $text_size);
  $color = rgb2cmyk(hex2rgb('#333333'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  
  // Output the text at the desired coordinates
  $p->fit_textline($text, $xoffset + $x, $yoffset  - $y - 1.0, $options);
}

function label_under($p, $text, $x, $y, $options = '') {
  // Set globals
  global $xoffset;
  global $yoffset;
  
  
  // Initialize text style
  $font = $p->load_font("Helvetica", "winansi", "");
  $p->setfont($font, DND_CHARACTER_SHEET_UNDERLINE_HEADER);
  $color = rgb2cmyk(hex2rgb('#999999'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  
  // Output the Label text at the desired coordinates
  $p->fit_textline(strtoupper($text), $xoffset + $x, $yoffset - DND_CHARACTER_SHEET_UNDERLINE_HEADER - $y - 2.0, $options);
}

function label_over($p, $text, $startx, $starty, $width, $options = "", $border = "") {
  // Set globals
  global $xoffset;
  global $yoffset;
  
  
  // Initialize text style
  $font = $p->load_font("Helvetica", "winansi", "");
  $p->setfont($font, DND_CHARACTER_SHEET_UNDERLINE_HEADER);
  $color = rgb2cmyk(hex2rgb('#333333'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  
  $options .= " position=center";
  $options .= " boxsize={{$width} 0}";
  
  // Output the Label text at the desired coordinates
  $p->fit_textline($text, $xoffset + $startx, $yoffset - DND_CHARACTER_SHEET_UNDERLINE_HEADER - $starty - 2.0, $options);
}

function label_middle($p, $text, $startx, $starty, $width, $options = "", $border = "") {
  // Set globals
  global $xoffset;
  global $yoffset;
  
  
  // Initialize text style
  $font = $p->load_font("Helvetica", "winansi", "");
  $p->setfont($font, DND_CHARACTER_SHEET_UNDERLINE_HEADER);
  $color = rgb2cmyk(hex2rgb('#333333'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  
  $options .= " position=center";
  $options .= " boxsize={{$width} 0}";
  
  // Output the Label text at the desired coordinates
  $p->fit_textline($text, $xoffset + $startx, $yoffset - DND_CHARACTER_SHEET_UNDERLINE_HEADER - $starty - 2.0, $options);
}

function label_filled($p, $text, $startx, $starty, $width, $options = "") {
  // Set globals
  global $xoffset;
  global $yoffset;
  
  
  // Initialize line style
  $color = rgb2cmyk(hex2rgb('#666666'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  $p->setlinewidth(0.2);
  $color = rgb2cmyk(hex2rgb('#eeeeee'));
  $p->setcolor("fill", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  $p->setlinewidth(0.2);
  
  // Draw the filled/stroked rectangle
  $p->moveto($xoffset + $startx, $yoffset - $starty);
  $p->lineto($xoffset + $startx + $width, $yoffset - $starty);
  $p->lineto($xoffset + $startx + $width, $yoffset - (DND_CHARACTER_SHEET_FILLED_HEADER * 2.0 )- $starty);
  $p->lineto($xoffset + $startx, $yoffset - (DND_CHARACTER_SHEET_FILLED_HEADER * 2.0 ) - $starty);
  $p->lineto($xoffset + $startx, $yoffset - $starty);
  $p->fill_stroke();

  
  // Initialize text style
  $font = $p->load_font("Helvetica", "winansi", "");
  $p->setfont($font, DND_CHARACTER_SHEET_FILLED_HEADER);
  $color = rgb2cmyk(hex2rgb('#333333'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  
  $options .= " position=center";
  $options .= " boxsize={{$width} 0}";

  // Output the Label text at the desired coordinates
  $p->fit_textline(strtoupper($text), $xoffset + $startx, $yoffset - DND_CHARACTER_SHEET_FILLED_HEADER - $starty, $options);
}
function label_skill($p, $text, $startx, $starty, $width, $options = "") {
  // Set globals
  global $xoffset;
  global $yoffset;
  
  
      
  // Initialize line style
  $color = rgb2cmyk(hex2rgb('#666666'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  $p->setlinewidth(0.2);
  $color = rgb2cmyk(hex2rgb('#eeeeee'));
  $p->setcolor("fill", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  $p->setlinewidth(0.2);
  
  // Draw the filled/stroked rectangle
  $p->moveto($xoffset + $startx, $yoffset - $starty);
  $p->lineto($xoffset + $startx + $width, $yoffset - $starty);
  $p->lineto($xoffset + $startx + $width, $yoffset - (DND_CHARACTER_SHEET_FILLED_HEADER * 2.0 )- $starty);
  $p->lineto($xoffset + $startx, $yoffset - (DND_CHARACTER_SHEET_FILLED_HEADER * 2.0 ) - $starty);
  $p->lineto($xoffset + $startx, $yoffset - $starty);
  $p->fill_stroke();

  // Initialize text style
  $font = $p->load_font("Helvetica", "winansi", "");
  $p->setfont($font, DND_CHARACTER_SHEET_UNDERLINE_HEADER);
  $color = rgb2cmyk(hex2rgb('#333333'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  
  $options .= " position={left center}";
  $options .= " boxsize={{$width} 0}";
  $options .= " margin={3 0}";

  // Output the Label text at the desired coordinates
  $p->fit_textline($text, $xoffset + $startx, $yoffset - DND_CHARACTER_SHEET_FILLED_HEADER - $starty, $options);
}

function cell($p, $text, $startx, $starty, $width, $options = "", $border = "", $size = '', $align='center') {
  // Set globals
  global $xoffset;
  global $yoffset;

  $text_size = ($size == 'small') ? DND_CHARACTER_SHEET_UNDERLINE_TEXT_SMALL : DND_CHARACTER_SHEET_UNDERLINE_TEXT;
  
  // Initialize line style
  if ($border == 'dashed') {
    $color = rgb2cmyk(hex2rgb('#666666'));
    $p->setdash(1, 2);
  } else {
    $color = rgb2cmyk(hex2rgb('#666666'));
  }
  
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  $p->setlinewidth(0.2);
  
  // Draw the filled/stroked rectangle
  $p->moveto($xoffset + $startx, $yoffset - $starty);
  $p->lineto($xoffset + $startx + $width, $yoffset - $starty);
  $p->lineto($xoffset + $startx + $width, $yoffset - (DND_CHARACTER_SHEET_FILLED_HEADER * 2.0 )- $starty);
  $p->lineto($xoffset + $startx, $yoffset - (DND_CHARACTER_SHEET_FILLED_HEADER * 2.0 ) - $starty);
  $p->lineto($xoffset + $startx, $yoffset - $starty);
  $p->stroke();
  $p->setdash(0, 0);
  
  // Initialize text style
  $font = $p->load_font("Helvetica-Bold", "winansi", "");
  $p->setfont($font, $text_size);
  $color = rgb2cmyk(hex2rgb('#333333'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  
  // Set text options for the cell
  $options .= " position=${align}";
  $options .= " boxsize={{$width} 0}";
  
  $text_variations = ($size == 'small') ? 2 : 0;
  $text_variations = ($align == 'left') ? 3 : $text_variations;
  $p->fit_textline($text, $xoffset + $startx + $text_variations, $yoffset - DND_CHARACTER_SHEET_FILLED_HEADER - $starty - $text_variations, $options);
}

function divider($p, $startx, $starty, $endx, $endy) {
  // Set globals
  global $xoffset;
  global $yoffset;
  
  
  $p->moveto($xoffset + $startx, $yoffset - $starty);
  $p->lineto($xoffset + $endx, $yoffset - $endy);
  $p->stroke();
}

function checkbox($p, $startx, $starty, $checked = '') {
  global $xoffset;
  global $yoffset;
  $size = 8.0;
  
  // Initialize line style
  $color = rgb2cmyk(hex2rgb('#666666'));
  $p->setcolor("both", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  $p->setlinewidth(0.2);
  $color = ($checked == 'checked') ? rgb2cmyk(hex2rgb('#111111')) : rgb2cmyk(hex2rgb('#eeeeee'));
  $p->setcolor("fill", "cmyk", $color['c'], $color['m'], $color['y'], $color['k']);
  $p->setlinewidth(0.2);
  
  // Draw the filled/stroked rectangle
  $p->moveto($xoffset + $startx, $yoffset - 4.0 - $starty);
  $p->lineto($xoffset + $startx + $size, $yoffset - 4.0 - $starty);
  $p->lineto($xoffset + $startx + $size, $yoffset - 4.0 - ($size)- $starty);
  $p->lineto($xoffset + $startx, $yoffset - 4.0 - ($size) - $starty);
  $p->lineto($xoffset + $startx, $yoffset - 4.0 - $starty);
  $p->fill_stroke();
}



function hex2rgb($hex) {
  $color = str_replace('#', '', $hex);
  $rgb = array(
    'r' => hexdec(substr($color,0,2)),
    'g' => hexdec(substr($color,2,2)),
    'b' => hexdec(substr($color,4,2)),
  );
  return $rgb;
}

function rgb2cmyk($var1, $g = 0, $b = 0) {
  if(is_array($var1)) {
    $r = $var1['r'];
    $g = $var1['g'];
    $b = $var1['b'];
  }
  else $r = $var1;
  $cyan    = 255 - $r;
  $magenta = 255 - $g;
  $yellow  = 255 - $b;
  $black   = min($cyan, $magenta, $yellow);
  $cyan    = @(($cyan    - $black) / (255 - $black)) * 255;
  $magenta = @(($magenta - $black) / (255 - $black)) * 255;
  $yellow  = @(($yellow  - $black) / (255 - $black)) * 255;
  
  return array(
    'c' => $cyan / 255,
    'm' => $magenta / 255,
    'y' => $yellow / 255,
    'k' => $black / 255,
  );
}






















// Row definitions
function define_rows($page) {

  $row = array(0.0);
  
  switch($page) {
    case 1:
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT; //1
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT; //5
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT; //10
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT + 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //15
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT + 8.0; //20
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //25
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT + 8.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //30
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT + 8.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //35
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT + 8.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //40
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //45
      break;
    case 2:
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT + 2.0; //1
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //5
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //10
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //15
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //20
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT + 8.0; //25
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //30
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //35
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //40
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //45
      break;
    case 3:
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //1
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //5
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //10
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //15
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //20
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //25
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //30
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //35
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //40
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //45
      break;
    case 4:
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //1
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //5
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //10
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //15
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT + 8.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //20
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //25
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //30
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0;
      $row[] = $row[count($row) - 1] + DND_CHARACTER_SHEET_LINE_HEIGHT + DND_CHARACTER_SHEET_HEADER_HEIGHT - 2.0; //35
      break;
  }
    
  return $row;
}

























function _get_traits_list($node) {
  $traits = array();
  $rawtraits = explode("\n", $node->character['race']['data']['traits_list']);
  foreach ($rawtraits as $trait) {
    $temp = explode("|||", wordwrap($trait, 50, '|||'));
    foreach ($temp as $key => $value) {
      $traits[] = ($key) ? '  ' . $value : $value;
    }
    
  }

  return $traits;
}




function _get_feats_list($node) {
  
  $feats = array();
  $rawfeats = $node->character['feats'];
  foreach ($rawfeats as $feat) {
    $benefit = $feat['benefit'];
    $benefit = str_replace('</p>', '</p> ', $benefit);
    $benefit = strip_tags($benefit);
    $benefit = str_replace('Benefit: ', '', $benefit);
    $benefit = trim($benefit);
    $feat_text  = $feat['name'] . ': ' . $benefit;

    $temp = explode("|||", wordwrap($feat_text, 133, '|||'));
    foreach ($temp as $key => $value) {
      $line = $value;
      $feats[] = ($key) ? '  ' . $line : $line;
    }

  }
  
/*
  print '<pre style="background:#fff;color:#000;font-family:consolas;font-size:16px;">';
  print_r($feats);
  print '<a name="bottom"></a>';
  die;
*/
  return $feats;
}







function _dnd_character_sheet_get_carry_load($node) {
  $capacities = array(
    array(0, 0, 0),
    array(3, 6, 10),
    array(6, 13, 20),
    array(10, 20, 30),
    array(13, 26, 40),
    array(16, 33, 50),
    array(20, 40, 60),
    array(23, 46, 70),
    array(26, 53, 80),
    array(30, 60, 90),
    array(33, 66, 100),
    array(38, 76, 115),
    array(43, 86, 130),
    array(50, 100, 150),
    array(58, 116, 175),
    array(66, 133, 200),
    array(76, 153, 230),
    array(86, 173, 260),
    array(100, 200, 300),
    array(116, 233, 350),
    array(133, 266, 400),
    array(153, 306, 460),
    array(173, 346, 520),
    array(200, 400, 600),
    array(233, 466, 700),
    array(266, 533, 800),
    array(306, 613, 920),
    array(346, 693, 1,040),
    array(400, 800, 1,200),
    array(466, 933, 1,400),
  );

  $slow = 20;
  $speed = (empty($node->field_dnd_character_speed[$node->language][0]['value'])) ? 30 : $node->field_dnd_character_speed[$node->language][0]['value'];
  switch ($node->field_dnd_character_speed[$node->language][0]['value']) {
    case 40:
      $slow = 30;
      break;
    case 30:
      $slow = 20;
      break;
    case 20:
      $slow = 15;
      break;
  }

  $carry  = array(
    'light' => array(
      'weight' => $capacities[$node->field_dnd_character_abilities[$node->language][0]['score']][0],
      'max_dex' => NULL,
      'check_penalty' => NULL,
      'speed' => $speed,
    ),
    'medium' => array(
      'weight' => $capacities[$node->field_dnd_character_abilities[$node->language][0]['score']][1],
      'max_dex' => 3,
      'check_penalty' => -3,
      'speed' => $slow,
    ),
    'heavy' => array(
      'weight' => $capacities[$node->field_dnd_character_abilities[$node->language][0]['score']][2],
      'max_dex' => 1,
      'check_penalty' => -6,
      'speed' => $slow,
    ),
  );
  
  return $carry;
}